"""
VARA Modem Listener.
"""
import argparse
import array
import logging
import time

import gensio

from .gutils import IOEvent, PipeEvent, OSFUNCS
from .rmsgw import RMSGatewayLogin


logger = logging.getLogger("vara")


WAIT_FOR_DISCONNECT = 120
WAIT_FOR_STEP_MSEC = 1000


class VaraPipeEvent(PipeEvent):
    def __init__(self, vara_control):
        super().__init__()
        self.vara_control = vara_control

    def connect_channel(self, spawn):
        self.reset_channel()
        # connect the program/console
        self.io2 = gensio.gensio(OSFUNCS, spawn, self)
        self.io2.open(self)

    def close_channel(self):
        self.close(self.io2)

    def reset_channel(self):
        # clean up any buffer remnants
        buf = self.bufA  # io2 writes from bufA
        buf[:] = buf[1:0]
        self.in_close = False

    def graceful_disconnect(self):
        self.vara_control.execute("DISCONNECT")
        # wait until VARA acknowledges the disconnect
        abort = time.time() + (WAIT_FOR_DISCONNECT * 0.9)
        expire = time.time() + WAIT_FOR_DISCONNECT
        while True:
            if time.time() > abort:
                self.vara_control.execute("ABORT")
                abort = float("inf")  # dont abort again
            if time.time() > expire:
                self.logger.error("VARA control never signaled disconnect.")
                break
            if not self.vara_control.connected:
                break
            self.waiter.wait_timeout(1, WAIT_FOR_STEP_MSEC)

    def close(self, io=None):
        if (io is not None and io.same_as(self.io)) or self.vara_control.in_shutdown:
            if self.vara_control.connected:
                try:
                    self.graceful_disconnect()
                except Exception as exc:
                    self.logger.exception("Unhandled Exception in graceful_disconnect")
            if not self.vara_control.in_shutdown:
                # Note: we do NOT want to close the VARA side of the data connection
                #       until we're shutting down the control side as well.
                return
        super().close(io)


class VaraRMSPipeEvent(RMSGatewayLogin, VaraPipeEvent):
    @property
    def callsign(self):
        return self.vara_control.laddr.partition("-")[0].encode("utf-8")


DEFAULT_SPAWN = "stdio(self)"


class VaraControlEvent(IOEvent):
    def __init__(self, laddr, data_port, spawn=None, banner=None, rms=False):
        super().__init__()
        self.laddr = laddr
        self.data_port = data_port
        self.spawn = spawn or DEFAULT_SPAWN
        self.banner = banner
        self.connected = None
        self.in_shutdown = False
        if rms:
            pipe = VaraRMSPipeEvent
        else:
            pipe = VaraPipeEvent
        self.data_pipe = self.establish_data_connection(pipe=pipe)
        # this will be sent on the first `write_callback`
        self.bufB[:] = array.array(
            "B",
            b"\r".join(
                [
                    f"MYCALL {laddr}".encode("utf-8"),
                    b"LISTEN ON\r",
                ],
            ),
        )

    @classmethod
    def from_gensio_str(cls, gensio_str, **kwargs):
        vc = cls(**kwargs)
        vc.io = gensio.gensio(OSFUNCS, gensio_str, vc)
        vc.io.open(vc)
        return vc

    def establish_data_connection(self, pipe=VaraPipeEvent):
        data_pipe = pipe(vara_control=self)
        data_pipe.io = gensio.gensio(
            OSFUNCS,
            self.data_port,
            data_pipe,
        )
        data_pipe.io.open(data_pipe)
        return data_pipe

    def dispatch(self, command):
        args = command.split()
        command = args[0].lower()
        if command == "connected":
            source, destination = args[1:3]
            if source == self.laddr:
                # we initiated a connection to destination
                self.connected = destination
            else:
                # we accepted a connection from source
                self.connected = source
            try:
                self.data_pipe.connect_channel(self.spawn)
            except Exception:
                self.log_for(self.io, f"Connection failed", exc_info=True)
                self.data_pipe.close_channel()
                return
            if self.banner is not None:
                self.data_pipe.bufB.extend(f"{self.banner}\r\n".encode("utf-8"))
            self.log_for(self.io, f"Connected: s:{source} -> d:{destination}")
        elif command == "disconnected":
            disconnected_from = self.connected
            self.connected = None
            if self.data_pipe is not None:
                self.data_pipe.reset_channel()
            self.logger.info(f"Disconnected from {disconnected_from}")
        elif command == "wrong":
            self.log_for(self.io, f"WRONG!", level=logging.INFO)
        elif command in [
            "ok",
            "iamalive",
            "ptt",
            "buffer",
            "registered",
            "link",
            "pending",
            "cancelpending",
            "busy",
        ]:
            # these commands are all ignored at this point
            pass
        else:
            self.log_for(
                self.io,
                f"unrecognized command: {command}",
                level=logging.INFO,
            )

    def execute(self, *args):
        def encode_utf8(obj):
            if not isinstance(obj, bytes):
                return str(obj).encode("utf-8")

        self.bufB.extend(
            b" ".join(encode_utf8(a) for a in args) + b"\r",
        )
        if self.io is not None:
            # schedule to send if we didn't already close
            self.io.write_cb_enable(True)

    def read_callback(self, io, err, data, auxdata):
        data_len = super().read_callback(io, err, data, auxdata)
        if err:
            return data_len
        if self.data_pipe.io is None:
            self.close()
            return data_len
        readBuf = self.get_read_buffer(io)
        if readBuf:
            readBuf[:], commands = (
                array.array("B"),
                readBuf.tobytes().decode("utf-8").strip().splitlines(),
            )
            for command in commands:
                try:
                    self.dispatch(command)
                except Exception as exc:
                    self.logger.exception("Unhandled Exception in dispatch")
        return data_len

    def close(self, io=None):
        self.in_shutdown = True
        self.data_pipe.close()
        super().close(io)

    def wait_till_close(self):
        while self.io is not None:
            try:
                self.wait()
            except KeyboardInterrupt:
                if self.in_close:
                    raise  # second interrupt kills us
                self.logger.info("Interrupt, closing modem...")
            finally:
                self.close()


"""
Some other ideas for spawning a gateway
SPAWN = "tcp,india.colorado.edu,13"   # time
SPAWN = "tcp,telehack.com,23"         # sort of shell
"""


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen",
        required=True,
        help="Callsign-SSID to listen for VARA connections.",
    )
    parser.add_argument(
        "-p",
        "--vara-port",
        default=8300,
        help="Control port for VARA or VARA FM",
    )
    parser.add_argument(
        "--vara-host",
        default="localhost",
        help="Host running VARA or VARA FM",
    )
    parser.add_argument(
        "--gateway",
        default="rms,tcp,cms.winlink.org,8772",
        help="Gensio connection string for gateway endpoint.",
    )
    parser.add_argument(
        "--banner",
        default=None,
        help="Text displayed when user connects",
    )

    args = parser.parse_args()

    gw1, rms, gw2 = args.gateway.partition("rms,")

    VaraControlEvent.from_gensio_str(
        gensio_str=f"tcp,{args.vara_host},{args.vara_port}",
        laddr=args.listen,
        data_port=f"tcp,{args.vara_host},{int(args.vara_port)+1}",
        spawn=gw1 + gw2,
        banner=args.banner,
        rms=bool(rms),
    ).wait_till_close()


if __name__ == "__main__":
    main()



"""Utilities for working with gensio python binding."""

import array
import logging

import gensio


# python-native logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("gutils")


class GensioLogger:
    """gensio internal log callback"""

    def gensio_log(self, level, log):
        logger.debug("***%s log: %s", level, log)


gensio.gensio_set_log_mask(gensio.GENSIO_LOG_MASK_ALL)
OSFUNCS = None


class IOEvent:
    """
    Buffered gensio wrapper.

    i = IOEvent()
    i.io = gensio.gensio(OSFUNCS, "tcp,localhost,3333", i)
    i.io.open(i)
    i.wait()
    """

    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)
        self._io = None
        # io reads into bufA
        self.bufA = array.array("B")
        # io writes from bufB
        self.bufB = array.array("B")
        self.in_close = False  # true if either gensio is down or going down
        self.in_error = False  # true if self._io had an error
        self.waiter = gensio.waiter(OSFUNCS)

    @property
    def io(self):
        return self._io

    @io.setter
    def io(self, value):
        self._io = value
        if self._io is not None:
            self._io.set_cbs(self)

    def name_for(self, io):
        if io.same_as(self.io):
            return "io"
        raise ValueError("Unknown io: {!r}".format(io))

    def log_for(self, io, message, *args, level=logging.DEBUG, **kwargs):
        """
        Log a message in the context of the given `io`.
        """
        self.logger.log(
            level, "{} {}".format(self.name_for(io), message), *args, **kwargs
        )

    def close(self, io=None):
        self.in_close = True
        if self.io is not None and not self.in_error:
            # with in_close=True, write_callback will close io after buffer is drained
            # (unless there is an error)
            self.io.write_cb_enable(True)
        # if the caller gave us an IO, close it now
        if io is not None:
            # this is the ONLY place to direcly call gensio.close!
            io.close(self)

    def get_read_buffer(self, io):
        if io.same_as(self.io):
            return self.bufA
        raise ValueError("Unknown io: {!r}".format(io))

    def read_callback(self, io, err, data, auxdata):
        if err:
            self.log_for(io, "read: %s", err)
            if "remote end closed connection" not in str(err).lower():
                self.log_for(io, "read error: %s", err)
            self.in_error = True
            self.close(io)
        if self.in_error:
            return 0
        if data:
            self.log_for(io, "read_callback: data=%r", data)
            self.get_read_buffer(io).extend(data)
        if self.bufB:
            self.io.write_cb_enable(True)
        return len(data)

    def get_write_buffer(self, io):
        if io.same_as(self.io):
            return self.bufB
        raise ValueError("Unknown io: {!r}".format(io))

    def write_callback(self, io):
        buf = self.get_write_buffer(io)
        if buf and not self.in_error:
            try:
                count = io.write(buf.tobytes(), None)
            except Exception as e:
                self.log_for(io, "write: %s (in_error=%s, in_close=%s, buf=%s)", e, self.in_error, self.in_close, buf)
                if "remote end closed connection" not in str(err).lower():
                    self.log_for(io, "write error: %s", e)
                self.in_error = True
                buf[:] = []  # reset the buffer here to avoid infinite loop on connection drop
                self.close(io)
                return
            self.log_for(io, "write_callback: data=%s", buf[:count].tobytes())
            buf[:] = buf[count:]
        else:
            io.write_cb_enable(False)
            if self.in_close:
                self.close(io)

    def open_done(self, io, err):
        if err:
            self.log_for(io, "open error: %s", err)
            if self.io is not None:
                self.close(io)
                # normally io would be reset in close_done, but that wont
                # get called if the gensio failed to open
                self.io = None
            return
        self.log_for(io, "Opened gensio: %s", io)
        io.write_cb_enable(True)
        io.read_cb_enable(True)

    def close_done(self, io, wake_when_closed=True):
        self.log_for(io, "closed gensio: %s", io)
        setattr(self, self.name_for(io), None)
        if self.io is None and wake_when_closed:
            self.waiter.wake()

    def wait(self):
        if self.io is not None:
            self.waiter.wait(1)


class PipeEvent(IOEvent):
    """
    Connect two gensios and pass data between them.

    p = PipeEvent()
    p.io = gensio.gensio(OSFUNCS, "tcp,localhost,3333", p)
    p.io.open(p)
    p.io2 = gensio.gensio(OSFUNCS, "stdio(self)", p)
    p.io2.open(p)
    p.wait()
    """

    def __init__(self):
        super().__init__()
        self._io2 = None

    @property
    def io2(self):
        return self._io2

    @io2.setter
    def io2(self, value):
        self._io2 = value
        if self._io2 is not None:
            self._io2.set_cbs(self)

    def name_for(self, io):
        if io.same_as(self.io2):
            return "io2"
        return super().name_for(io)

    def close(self, io=None):
        super().close(io)
        if self.io2 is not None and not self.in_error:
            self.io2.write_cb_enable(True)

    def get_read_buffer(self, io):
        if io.same_as(self.io2):
            return self.bufB
        return super().get_read_buffer(io)

    def read_callback(self, io, err, data, auxdata):
        len_data = super().read_callback(io, err, data, auxdata)
        if self.bufA:
            self.io2.write_cb_enable(True)
        return len_data

    def get_write_buffer(self, io):
        if io.same_as(self.io2):
            return self.bufA
        return super().get_write_buffer(io)

    def open_done(self, io, err):
        super().open_done(io, err)
        if err and self.io2 is not None:
            self.close(self.io2)
            self.io2 = None
            return

    def close_done(self, io):
        super().close_done(io, wake_when_closed=False)
        if self.io is None and self.io2 is None:
            self.waiter.wake()

    def wait(self):
        if self.io is not None or self.io2 is not None:
            self.waiter.wait(1)


class ListenerEvent:
    """
    Listen and accept incoming connections.

    listener = ListenerEvent()
    listener.acc = gensio.gensio_accepter(OSFUNCS, "tcp,3333", listener)
    listener.acc.startup()
    listener.wait()
    """

    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)
        self.acc = None
        self.ios = []
        self.waiter = gensio.waiter(OSFUNCS)
        self.in_shutdown = False

    def log(self, acc, level, logval):
        self.logger.error("gensio acc %s err: %s", level, logval)

    def new_connection(self, acc, io):
        if self.in_shutdown:
            # it will free automatically
            return None
        self.ios.append(io)
        return io

    def check_finish(self):
        if len(self.ios) == 0 and self.acc is None:
            self.waiter.wake()

    def io_closed(self, io):
        """
        Called from gensio `close_done` callback to signal to listener
        that the connection is closed and no longer needs to be tracked.
        """
        for idx, saved_io in enumerate(self.ios):
            if saved_io.same_as(io):
                break
        else:
            self.logger.warning("untracked connection closed: %r", io)
            idx = None
        if idx is not None:
            del self.ios[idx]
        self.check_finish()

    def shutdown_done(self, acc):
        self.acc = None
        self.check_finish()

    def shutdown(self):
        if self.in_shutdown:
            return
        self.in_shutdown = True
        self.acc.shutdown(self)
        for i in self.ios:
            i.close_s()

    def wait(self):
        if self.acc is not None:
            self.waiter.wait(1)


INIT_DONE = False


def init():
    global INIT_DONE
    if INIT_DONE:
        return
    global OSFUNCS
    OSFUNCS = gensio.alloc_gensio_selector(GensioLogger())
    INIT_DONE = True


init()



"""
Mixin for handling Winlink RMS Gateway login.

For use with AX.25 or VARA winlink gateways.
"""


class RMSGatewayLogin:
    """For best results, mix with gutils.PipeEvent."""

    def _handle_login(self, io, err, data):
        if (
            getattr(self, "_cms_logged_in", False)
            or not data
            or err
            or not io.same_as(self.rms_io)
        ):
            return

        ldata = data.strip().lower()
        send_line = None
        if ldata.startswith(b"callsign"):
            send_line = b"%s\r\n" % self.callsign
        elif ldata.startswith(b"password"):
            send_line = b"%s\r\n" % self.password
            self._cms_logged_in = True
        if send_line is not None:
            self.logger.info("_handle_login: data={}".format(ldata))
            self.get_write_buffer(io).extend(send_line)
            io.write_cb_enable(True)
            return True

    def read_callback(self, io, err, data, auxdata):
        if self._handle_login(io, err, data):
            return len(data)
        return super().read_callback(io, err, data, auxdata)

    @property
    def callsign(self):
        return b"N0CALL"

    @property
    def password(self):
        return b"CMSTelnet"

    @property
    def rms_io(self):
        return self.io2

    def close(self, io=None):
        if io and io.same_as(self.rms_io):
            self._cms_logged_in = False



import gensio

from gensio_modems.gutils import ListenerEvent, OSFUNCS, PipeEvent


class StdioPipeListener(ListenerEvent):
    def new_connection(self, acc, io):
        pev = PipeEvent()
        pev.io = super().new_connection(acc, io)
        pev.io.write_cb_enable(True)
        pev.io.read_cb_enable(True)
        pev.io2 = gensio.gensio(OSFUNCS, "stdio(self)", pev)
        pev.io2.open(pev)


listener = StdioPipeListener()
listener.acc = gensio.gensio_accepter(OSFUNCS, "tcp,3334", listener)
listener.acc.startup()
listener.wait()



"""AX.25 / KISS modem implementation."""
import argparse
import logging

import gensio

from gensio_modems.gutils import ListenerEvent, PipeEvent, OSFUNCS
from gensio_modems.rmsgw import RMSGatewayLogin


logger = logging.getLogger("ax25")


def ax25_addr(io, control):
    addr = io.control(0, True, control, b"")
    gensio_name, _, addr_str = addr.partition(":")
    addrs = addr_str.split(",")
    return addrs[1] if len(addrs) > 1 else addrs[0]


def ax25_remote_addr(io):
    return ax25_addr(io, control=gensio.GENSIO_CONTROL_RADDR)


def ax25_local_addr(io):
    return ax25_addr(io, control=gensio.GENSIO_CONTROL_LADDR)


def ax25_command_substitution(io, command):
    remote_addr = ax25_remote_addr(io)
    remote_callsign = remote_addr.partition("-")[0]
    # from https://manpages.ubuntu.com/manpages/jammy/man5/ax25d.conf.5.html
    replmap = {
        "%d": "gax25",
        "%U": remote_callsign.upper(),
        "%u": remote_callsign.lower(),
        "%S": remote_addr.upper(),
        "%s": remote_addr.lower(),
        "%P": "%%",
        "%p": "%%",
        "%R": "%%",
        "%r": "%%",
        "%%": "%%",
    }
    for find, repl in replmap.items():
        if find in command:
            command = command.replace(find, repl)
    return command


def spawn_for(ioev, gensio_str):
    sh_cmd = ax25_command_substitution(ioev.io, gensio_str)
    logger.info("spawn: {}".format(sh_cmd))
    return gensio.gensio(OSFUNCS, sh_cmd, ioev)


class RMSPipeEvent(RMSGatewayLogin, PipeEvent):
    @property
    def callsign(self):
        return ax25_remote_addr(self.io).partition("-")[0].encode("utf-8")


class AX25ListenerEvent(ListenerEvent):
    def __init__(self, spawn_gensio_str, banner=None):
        gw1, rms, gw2 = spawn_gensio_str.partition("rms,")
        self.spawn_gensio_str = gw1 + gw2
        self.banner = banner
        self.pipe = RMSPipeEvent if rms else PipeEvent
        super().__init__()

    @classmethod
    def from_gensio_str(cls, gensio_str, **kwargs):
        accev = cls(**kwargs)
        accev.acc = gensio.gensio_accepter(OSFUNCS, gensio_str, accev)
        accev.acc.startup()
        return accev

    def new_connection(self, acc, io):
        opened_io = super().new_connection(acc, io)
        if opened_io is None:
            return None
        ioev = self.pipe()
        ioev.io = io
        if self.banner is not None:
            ioev.get_write_buffer(ioev.io).extend(f"{self.banner}\r\n".encode("utf-8"))
        ioev.io2 = spawn_for(ioev, self.spawn_gensio_str)
        ioev.io2.open(ioev)
        io.write_cb_enable(True)
        io.read_cb_enable(True)
        return io


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen",
        required=True,
        help="Callsign-SSID to listen for AX25 connections.",
    )
    parser.add_argument(
        "-k",
        "--kiss",
        default="tcp,localhost,8001",
        help="Gensio kiss string",
    )
    parser.add_argument(
        "-g",
        "--gateway",
        default="rms,tcp,cms.winlink.org,8772",
        help="Gensio connection string for gateway endpoint.",
    )
    parser.add_argument(
        "--banner",
        default=None,
        help="Text displayed when user connects",
    )

    args = parser.parse_args()

    AX25ListenerEvent.from_gensio_str(
        gensio_str=f"ax25(laddr={args.listen},extended=0),kiss,conacc,{args.kiss}",
        spawn_gensio_str=args.gateway,
        banner=args.banner,
    ).wait()


if __name__ == "__main__":
    main()



import gensio

from gensio_modems.gutils import PipeEvent, OSFUNCS


p = PipeEvent()
p.io = gensio.gensio(OSFUNCS, "tcp,localhost,3333", p)
p.io.open(p)
p.io2 = gensio.gensio(OSFUNCS, "stdio(self)", p)
p.io2.open(p)
p.wait()



#!/usr/bin/python3
"""
gAXproxy - proxy an AX.25 connection over telnet.

use with pat to establish a "telnet" connection that uses gensio for
AX.25 transport.
"""

import array
import argparse
import logging
import re
import shlex
import subprocess
import sys

import gensio


# python-native logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("gaxproxy")

CRED_PROMPTS = [b"Callsign :\r", b"Password :\r"]


class GensioLogger:
    """gensio internal log callback"""

    def gensio_log(self, level, log):
        logger.debug("***%s log: %s", level, log)


gensio.gensio_set_log_mask(gensio.GENSIO_LOG_MASK_ALL)
o = gensio.alloc_gensio_selector(GensioLogger())


def replace_positional(endpoint, *args):
    for ix, val in enumerate(args):
        endpoint = endpoint.replace(f"%{ix}", val)
    return endpoint


def spawn_for(ioev, endpoint):
    logger.info(f"spawning {endpoint}")
    return gensio.gensio(o, endpoint, ioev)


class IOEvent:
    def __init__(self, accev, require_creds=1):
        self.accev = accev
        self._io = None  # the "listener" end, typically telnet connection
        self._io2 = None  # established on demand to make an ax25 connection
        # io reads into bufA and io2 writes from bufA
        self.bufA = array.array("B")
        # io writes from bufB and io2 reads into bufB
        self.bufB = array.array("B")
        self.in_close = False  # true if either gensio is down or going down
        self.require_creds = require_creds
        self.creds = []
        if self.require_creds == 1:
            # the "Password"-only style
            self.bufB.extend(CRED_PROMPTS[1])
        elif self.require_creds:
            self.bufB.extend(CRED_PROMPTS[0])

    @property
    def io(self):
        return self._io

    @io.setter
    def io(self, value):
        self._io = value
        if self._io is not None:
            self._io.set_cbs(self)
            self.spawn_io2_when_ready()

    @property
    def io2(self):
        return self._io2

    @io2.setter
    def io2(self, value):
        self._io2 = value
        if self._io2 is not None:
            self._io2.set_cbs(self)

    def name_for(self, io):
        if io.same_as(self.io):
            return "io"
        elif io.same_as(self.io2):
            return "io2"
        raise ValueError("Unknown io: {!r}".format(io))

    def log_for(self, io, message, *args, level=logging.DEBUG, **kwargs):
        """
        Log a message in the context of the given `io`.
        """
        logger.log(level, "{} {}".format(self.name_for(io), message), *args, **kwargs)

    def close(self, io):
        self.in_close = True
        # with in_close=True, write_callback will close io after buffer is drained
        self.io.write_cb_enable(True)
        self.io2.write_cb_enable(True)
        io.close(self)

    def get_read_buffer(self, io):
        if io.same_as(self.io):
            return self.bufA
        elif io.same_as(self.io2):
            return self.bufB
        raise ValueError("Unknown io: {!r}".format(io))

    def spawn_io2_when_ready(self):
        if (
            self.io2 is None
            and len(self.creds) >= self.require_creds
            and not self.in_close
        ):
            # once creds are received, spawn and connect the second endpoint
            self.io2 = spawn_for(
                ioev=self,
                endpoint=replace_positional(self.accev.endpoint, *self.creds),
            )
            self.io2.open(self)

    def read_callback(self, io, err, data, auxdata):
        if err:
            self.log_for(io, "read: %s", err)
            if err != "Remote end closed connection":
                self.log_for(io, "read error: %s", err)
            self.close(io)
            return 0
        if data:
            self.log_for(io, "read_callback: data=%r", data)
            if len(self.creds) < self.require_creds and io.same_as(self.io):
                self.creds.append(data.decode("utf-8").strip())
                if len(self.creds) < self.require_creds:
                    self.bufB.extend(CRED_PROMPTS[len(self.creds) % len(CRED_PROMPTS)])
            else:
                self.get_read_buffer(io).extend(data)
        self.spawn_io2_when_ready()
        if self.bufA:
            self.io2.write_cb_enable(True)
        if self.bufB:
            self.io.write_cb_enable(True)
        return len(data)

    def get_write_buffer(self, io):
        if io.same_as(self.io):
            return self.bufB
        elif io.same_as(self.io2):
            return self.bufA
        raise ValueError("Unknown io: {!r}".format(io))

    def write_callback(self, io):
        buf = self.get_write_buffer(io)
        if buf:
            try:
                count = io.write(buf.tobytes(), None)
            except Exception as e:
                self.log_for(io, "write: %s", e)
                if str(e) != "Remote end closed connection":
                    self.log_for(io, "write error: %s", e)
                self.close(io)
                return
            self.log_for(io, "write_callback: data=%s", buf[:count].tobytes())
            buf[:] = buf[count:]

        if not buf:
            io.write_cb_enable(False)
            if self.in_close:
                io.close(self)

    def open_done(self, io, err):
        if err:
            self.log_for(io, "open error: %s", err)
            for io in [self.io, self.io2]:
                if io is not None:
                    try:
                        io.close(self)
                    except Exception:
                        pass  # ignore close errors for unestablished sessions
            return
        self.log_for(io, "Opened gensio: %s", io)
        io.write_cb_enable(True)
        io.read_cb_enable(True)

    def close_done(self, io):
        self.log_for(io, "closed gensio: %s", io)
        setattr(self, self.name_for(io), None)
        if self.io is None and self.io2 is None:
            self.accev.io_closed(self)

            # Break reference loops
            self.accev = None


class AccEvent:
    def __init__(self, endpoint, require_creds):
        self.ios = []
        self.waiter = gensio.waiter(o)
        self.endpoint = endpoint
        self.require_creds = require_creds
        self.in_shutdown = False

    def log(self, acc, level, logval):
        logger.error("gensio acc %s err: %s", level, logval)

    def new_connection(self, acc, io):
        if self.in_shutdown:
            # it will free automatically
            return
        ioev = IOEvent(self, require_creds=self.require_creds)
        logger.info("accepted new connection: %r", io)
        ioev.io = io
        self.ios.append(ioev)
        # enable callbacks explicitly, since io is already open so
        # IOEvent.open_done will not be called
        io.write_cb_enable(True)
        io.read_cb_enable(True)

    def check_finish(self):
        if len(self.ios) == 0 and self.acc is None:
            self.waiter.wake()

    def io_closed(self, ioev):
        i = self.ios.index(ioev)
        del self.ios[i]
        logger.info("connection closed: %r", ioev)
        self.check_finish()

    def shutdown_done(self, acc):
        self.acc = None
        self.check_finish()

    def shutdown(self, ioev):
        if self.in_shutdown:
            return
        self.in_shutdown = True
        self.acc.shutdown(self)
        for i in self.ios:
            if i != ioev:
                # The caller will close itself, let it finish its write
                i.close(io)

    def wait(self):
        self.waiter.wait(1)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--mycall", help="my callsign")
    parser.add_argument(
        "--ax25-conf",
        default="extended=0,retries=2",
        help="parameters passed to gensio ax25 module",
    )
    parser.add_argument(
        "-l", "--listen", default="tcp,localhost,8772", help="gensio spec to listen on"
    )
    parser.add_argument(
        "--require-creds",
        default=1,
        help="Prompt for 'credentials' on listener (can be replaced as %%n in --gateway",
    )
    parser.add_argument(
        "-g",
        "--gateway",
        default="%0",
        help="gateway callsign to connect to; defaults to the provided 'password'",
    )
    parser.add_argument(
        "-k",
        "--kiss",
        default="tcp,localhost,8001",
        help="gensio kiss connection string",
    )
    args = parser.parse_args()

    endpoint_conf = args.ax25_conf
    if not re.search(r"\bladder=", endpoint_conf):
        endpoint_conf = f"laddr={args.mycall}," + endpoint_conf
    if not re.search(r"\badder=", endpoint_conf):
        endpoint_conf = f'addr="0,{args.gateway},{args.mycall}",' + endpoint_conf

    accev = AccEvent(
        endpoint=f"ax25({endpoint_conf}),kiss,{args.kiss}",
        require_creds=int(args.require_creds),
    )
    accev.acc = gensio.gensio_accepter(o, args.listen, accev)
    accev.acc.startup()
    accev.wait()


if __name__ == "__main__":
    main()
